"""
pytvheadend.constants
~~~~~~~~~~~~~~~~~~~~
Constants list
Copyright (c) 2019 John Mihalic <https://github.com/mezz64>
Licensed under the MIT license.
"""

MAJOR_VERSION = 0
MINOR_VERSION = 0
SUB_MINOR_VERSION = 1
__version__ = '{}.{}.{}'.format(
    MAJOR_VERSION, MINOR_VERSION, SUB_MINOR_VERSION)

# API_URL = 'https://app-api.8slp.net/v1'

SUBSCRIPTIONS_URL = '/api/status/subscriptions'
CHANNELS_URL = '/api/channel/grid?start=0&limit=999999999'
SERVICES_URL = '/api/mpegts/service/grid?start=0&limit=999999999'
MUXES_URL = '/api/mpegts/mux/grid?start=0&limit=999999999'

DEFAULT_PORT = 9981

DEFAULT_TIMEOUT = 60

DEFAULT_HEADERS = {
    # 'content-type': "application/x-www-form-urlencoded",
    # 'connection': "keep-alive",
    # 'accept': "*/*",
    }



"""
pytvheadend.stream
~~~~~~~~~~~~~~~~~~~~
Provides stream data for TVHeadend
Copyright (c) 2019 John Mihalic <https://github.com/mezz64>
Licensed under the MIT license.

"""
import asyncio
import json
import logging

_LOGGER = logging.getLogger(__name__)


class Stream(object):
    """TVHeadend stream object"""
    def __init__(self, server):
        """Initialize stream object."""
        self.server = server
        self._channel_name = None
        self._active_service = None

        self._service_list = []
        self._service_history = []

    @property
    def is_active(self):
        """Return if active or not"""
        return bool(self._channel_name)

    @property
    def channel_name(self):
        """Return channel name"""
        return self._channel_name

    @property
    def active_service(self):
        """Return currently active service"""
        return self._active_service

    @property
    def service_full_list(self):
        """Return list of services"""
        return self._service_list

    @property
    def service_name_list(self):
        """Return list of service names"""
        tmp_list = []
        for service in self._service_list:
            tmp_list.append(service['name'])
        return tmp_list

    def update_data(self, channel=None):
        """ Update subscription object. """
        if not channel:
            self._channel_name = None
            self._active_service = None
            self._service_list = []
            self._service_history = []
            _LOGGER.debug('Stream object cleared.')
        else:
            self._channel_name = channel['name']
            if channel['network'] != self._active_service:
                self._active_service = channel['network']
                self._service_history.append(self._active_service)
                self.get_channel_info()

            _LOGGER.debug('Channel updated: %s', self._channel_name)

    def get_channel_info(self):
        """Return list of services & muxes based on channel name"""
        options = []
        if not self._channel_name:
            return

        if not self.server.chan_json:
            return

        for chan in self.server.chan_json:
            if chan['name'].upper() == self._channel_name:
                for serv in chan['services']:
                    for mux in self.server.serv_json:
                        if mux['uuid'] == serv:
                            if mux['network'].upper() == self._active_service:
                                active = True
                            else:
                                active = False
                            options.append({
                                'name': mux['network'].upper(),
                                'service_uuid': mux['uuid'],
                                'mux_uuid': mux['multiplex_uuid'],
                                'active': active,
                                })
        self._service_list = options

    async def change_service(self, new_service=None):
        """Change active service"""
        disable_list = []
        enable_list = []
        service_valid = False

        if not self._channel_name:
            return

        # If no service is defined:
        # - Disable active and previously active services
        # - Wait 5s for switch, re-enable services
        if new_service:
            _LOGGER.debug('Changing to %s', new_service)
            # Disable all services except new one
            for network in self._service_list:
                if network['name'] == new_service:
                    service_valid = True
                    _LOGGER.debug('Found Service')

            if service_valid:
                for network in self._service_list:
                    if network['name'] != new_service:
                        disable_list.append({
                            "enabled": "false",
                            "uuid": network['service_uuid']
                            })
                        enable_list.append({
                            "enabled": "true",
                            "uuid": network['service_uuid']
                            })
        else:
            # Disable those we've already tried to get a new option
            # unless we've tried them all, if so reset the list
            if len(self._service_history) == len(self._service_list):
                # We've tried them all, reset history
                _LOGGER.debug('History full, being reset.')
                self._service_history = [self._active_service]

            for service in self._service_history:
                # Build json with the service uuids to disable.
                for network in self._service_list:
                    if service == network['name']:
                        disable_list.append({
                            "enabled": "false",
                            "uuid": network['service_uuid']
                            })
                        enable_list.append({
                            "enabled": "true",
                            "uuid": network['service_uuid']
                            })
        _LOGGER.debug(json.dumps(disable_list))

        # http://192.168.11.5:9981/api/idnode/save
        data = {'node': json.dumps(disable_list)}
        req = await self.server.api_post(
            self.server.root_url + "/api/idnode/save", params=None, data=data)
        if req is None:
            _LOGGER.error('Unable to disable services.')
        else:
            # Wait for service to switch
            await asyncio.sleep(2)
            # Renable all options
            data = {'node': json.dumps(enable_list)}
            req = await self.server.api_post(
                self.server.root_url + "/api/idnode/save", params=None, data=data)

        # Force update after we make a change
        await self.server.fetch_subscription_list(force=True)



# Copyright (c) 2017-2018 John Mihalic <https://github.com/mezz64>
# Licensed under the MIT license.

# Used this guide to create module
# http://peterdowns.com/posts/first-time-with-pypi.html

# git tag 0.1 -m "0.1 release"
# git push --tags origin master
#
# Upload to PyPI Live
# python setup.py register -r pypi
# python setup.py sdist upload -r pypi


from distutils.core import setup
setup(
    name='pyTVHeadend',
    packages=['pytvheadend'],
    version='0.0.1',
    description='Provides a python api to interact with a TVHeadend server.',
    author='John Mihalic',
    author_email='mezz64@users.noreply.github.com',
    url='https://github.com/mezz64/pyTVHeadend',
    download_url='https://github.com/mezz64/pytvheadend/tarball/0.0.1',
    keywords=['tv', 'tvheadend', 'api wrapper', 'homeassistant'],
    classifiers=[],
    )



"""
pytvheadend.tvheadend
~~~~~~~~~~~~~~~~~~~~
Provides api for TVHeadend
Copyright (c) 2019 John Mihalic <https://github.com/mezz64>
Licensed under the MIT license.

"""

import logging
import asyncio
import aiohttp
import async_timeout

from pytvheadend.stream import Stream
from pytvheadend.constants import (
    DEFAULT_TIMEOUT, DEFAULT_HEADERS, DEFAULT_PORT,
    SUBSCRIPTIONS_URL, CHANNELS_URL, SERVICES_URL,
    __version__)

_LOGGER = logging.getLogger(__name__)


# Workflow
# Request active subscriptions, keep dictionary of active channel names
# Associate channel name with services by requesting channel grid
# Associate service with mux by requesting service grid
# At this point we have whole chain and can enable/disable services to change the source

# Channel dictionary
# {
#  'id': chann['id'],
# 'name': chann['channel'].upper(),
# 'network': chann['service'].split("/")[1].upper(),
# }

# Easiest way to make this work well with HASS which doesn't
# really support the dynamic add/remove of entities:
# - Define a fixed number of external objects representing total active streams
# - Internally manage stream objects being connected to those objects
# - External objects have a value of either "Inactive" or a given stream

class TVHeadend(object):
    """TVHeadend API object."""
    def __init__(self, host=None, port=DEFAULT_PORT,
                 usr=None, pwd=None, maxconn=1, loop=None):
        """Initialize eight sleep class."""

        _LOGGER.debug("pyTVHeadend %s initializing new server at: %s",
                      __version__, host)

        if not host:
            _LOGGER.error('Host not specified! Cannot continue.')
            return

        self.host = host
        self.usr = usr
        self.pwd = pwd

        self.root_url = 'http://{}:{}'.format(host, port)

        self.chan_json = None
        self.serv_json = None

        self._active_list = []

        self._streams = {}
        self._active_subscriptions = []

        self._ext_list = [None] * int(maxconn)
        for xxx in range(int(maxconn)):
            self._ext_list[xxx] = Stream(self)

        if loop is None:
            _LOGGER.info("Must supply asyncio loop.  Quitting")
            return None
        else:
            self._event_loop = loop
            self._own_loop = False

        asyncio.set_event_loop(self._event_loop)

        self._api_session = aiohttp.ClientSession(
            headers=DEFAULT_HEADERS, loop=self._event_loop)

        # Callbacks
        self._update_callbacks = []

    @property
    def active_subscriptions(self):
        """Return list of current subscriptions."""
        return self._active_subscriptions

    @property
    def active_streams(self):
        """Return dictionary of active streams"""
        return self._streams

    @property
    def stream_list(self):
        """Return external stream list"""
        return self._ext_list

    def add_update_callback(self, callback):
        """Register as callback for when a stream changes."""
        self._update_callbacks.append(callback)
        _LOGGER.debug('Added update callback to %s', callback)

    def remove_update_callback(self, callback):
        """ Remove a registered update callback. """
        if callback in self._update_callbacks:
            self._update_callbacks.remove(callback)
            _LOGGER.debug('Removed update callback %s',
                          callback)

    def _do_update_callback(self, msg):
        """Call registered callback functions."""
        for callback in self._update_callbacks:
            _LOGGER.debug('Update callback %s by %s',
                          callback, msg)
            self._event_loop.call_soon(callback, msg)

    async def start(self):
        """Start api initialization."""

        await self.fetch_channel_list()
        await self.fetch_service_list()
        return True

    async def stop(self):
        """Stop api session."""
        _LOGGER.debug('Closing tvheadend session.')
        await self._api_session.close()

    async def fetch_subscription_list(self, force=False):
        """Fetch list of active stream subscriptions"""
        # url = '{}/users/me'.format(API_URL)
        streams = []

        slist = await self.api_get(self.root_url + SUBSCRIPTIONS_URL,
                                   {'start': '0', 'limit': '999999999'})
        if slist is None:
            _LOGGER.error('Unable to fetch subscriptions.')
        else:
            # self._devices = dlist['user']['devices']
            # _LOGGER.debug('RAW: %s', slist)
            # _LOGGER.debug('RAW: %s', slist['entries'])
            for chann in slist['entries']:
                streams.append({
                    'id': chann['id'],
                    'name': chann['channel'].upper(),
                    'network': chann['service'].split("/")[1].upper(),
                    })

        self._active_subscriptions = streams
        # _LOGGER.debug(streams)
        self.update_stream_list(streams, force)
        log_str = ""
        for index, obj in enumerate(self._ext_list):
            log_str = log_str + 'Pos: {} - Name: {}, '.format(index, obj.channel_name)

# [{
#   'channel': 'Channel Name',
#   'networks': [{},{},...]
# },...]

# Use dictionary to keep track of channel in external list
#  {ext_list_index: stream_name}
#
#   Loop through active channel list, see if each channel is in the dictionary
#       If not:  Add it - (helper function to get highest empty index)
#       If so: update

    def next_index(self):
        """Return next available index to assign stream"""
        for index, strm in enumerate(self._ext_list):
            if not strm.channel_name:
                # Index is free
                # _LOGGER.debug('Returning available index: %s', index)
                return index

    def update_stream_list(self, streams, force):
        """ Update device list. """
        if streams is None:
            _LOGGER.error('Error updating TVHeadend streams, no data.')
            return

        active_streams = []
        for channel in streams:
            stream_name = '{}.{}'.format(channel['id'], channel['name'])
            active_streams.append(stream_name)

            if stream_name not in self._streams:
                _LOGGER.debug('New stream: %s. Adding to stream list.',
                              stream_name)
                index = self.next_index()
                self._streams[stream_name] = index
                _LOGGER.debug(self._streams)
                # new = Stream(self)
                # self._streams[stream_name] = new
                # new_streams.append(new)

            # Update no matter what
            self._ext_list[self._streams[stream_name]].update_data(channel)
            if force:
                self._do_update_callback(self._streams[stream_name])
            # self._streams[stream_name].update_data(channel)

        # Need to check for no longer active streams and remove
        tmp_strm = self._streams.copy()
        try:
            for stream, index in tmp_strm.items():
                # _LOGGER.debug('1Stream index: %s, Channel: %s', index, stream)
                if stream not in active_streams:
                    # stream no longer active
                    _LOGGER.debug('Old stream: %s. Removing from stream dict.',
                                  stream)
                    # self._streams.pop(stream_id)
                    del self._streams[stream]
                    self._ext_list[index].update_data()
                    # _LOGGER.debug(self._streams)
        except Exception as err:
            # This is a bad idea
            _LOGGER.debug('Caught: %s', err)

    async def fetch_channel_list(self):
        """Fetch channel list"""
        result = await self.api_get(self.root_url + CHANNELS_URL, {'start': '0', 'limit': '999999999'})
        if result is None:
            _LOGGER.error('Unable to fetch channels.')
        else:
            self.chan_json = result['entries']
            # _LOGGER.debug(result)

    async def fetch_service_list(self):
        """Fetch service list"""
        result = await self.api_get(self.root_url + SERVICES_URL, {'start': '0', 'limit': '999999999'})
        if result is None:
            _LOGGER.error('Unable to fetch services.')
        else:
            self.serv_json = result['entries']
            # _LOGGER.debug(result)

    def get_services(self, channel_name):
        """Return list of service IDs based on channel name"""
        if not self.chan_json:
            return

        for chan in self.chan_json:
            if chan['name'] == channel_name:
                return chan['services']

    async def api_post(self, url, params=None, data=None):
        """Make api post request."""
        post = None
        try:
            with async_timeout.timeout(DEFAULT_TIMEOUT, loop=self._event_loop):
                post = await self._api_session.post(
                    url, params=params, data=data)
            if post.status != 200:
                _LOGGER.error('Error posting data: %s', post.status)
                return None

            if 'text/x-json' in post.headers['content-type']:
                post_result = await post.json(content_type='text/x-json')
            else:
                _LOGGER.debug('Response was not JSON, returning text.')
                post_result = await post.text()

            return post_result

        except (aiohttp.ClientError, asyncio.TimeoutError,
                ConnectionRefusedError) as err:
            _LOGGER.error('Error posting data. %s', err)
            return None

    async def api_get(self, url, params=None):
        """Make api fetch request."""
        request = None
        headers = DEFAULT_HEADERS.copy()
        # headers.update({'Session-Token': self._token})

        try:
            with async_timeout.timeout(DEFAULT_TIMEOUT, loop=self._event_loop):
                request = await self._api_session.get(
                    url, headers=headers, params=params)
            # _LOGGER.debug('Get URL: %s', request.url)
            if request.status != 200:
                _LOGGER.error('Error fetching data: %s', request.status)
                return None

            if 'text/x-json' in request.headers['content-type']:
                request_json = await request.json(content_type='text/x-json')
            else:
                _LOGGER.debug('Response was not JSON, returning text.')
                request_json = await request.text()

            return request_json

        except (aiohttp.ClientError, asyncio.TimeoutError,
                ConnectionRefusedError) as err:
            _LOGGER.error('Error fetching data. %s', err)
            return None

    async def api_put(self, url, data=None):
        """Make api put request."""
        put = None
        headers = DEFAULT_HEADERS.copy()
        # headers.update({'Session-Token': self._token})

        try:
            with async_timeout.timeout(DEFAULT_TIMEOUT, loop=self._event_loop):
                put = await self._api_session.put(
                    url, headers=headers, data=data)
            if put.status != 200:
                _LOGGER.error('Error putting data: %s', put.status)
                return None

            if 'text/x-json' in put.headers['content-type']:
                put_result = await put.json(content_type='text/x-json')
            else:
                _LOGGER.debug('Response was not JSON, returning text.')
                put_result = await put.text()

            return put_result

        except (aiohttp.ClientError, asyncio.TimeoutError,
                ConnectionRefusedError) as err:
            _LOGGER.error('Error putting data. %s', err)
            return None
